@@ -94,11 +94,12 @@ static int fadvise_mode =
 #endif
   ;
 
-/* cache mode */
-static enum { cache_default, cache_none } cache_mode = cache_default;
+/* Reduce memory pressure. */
+static bool reduce_memory_pressure = false;
 
 /* Define EVICT_WRITES if we are going to evict the page cache
- * (cache=none) after writing.  This is only known to work on Linux.
+ * (reduce-memory-pressure=true) after writing.  This is only known to
+ * work on Linux.
  */
 #ifdef __linux__
 #define EVICT_WRITES 1
@@ -117,11 +118,13 @@ struct write_window {
 };
 
 static pthread_mutex_t window_lock = PTHREAD_MUTEX_INITIALIZER;
+static pthread_mutex_t window_fd_lock = PTHREAD_MUTEX_INITIALIZER;
 static struct write_window window[NR_WINDOWS];
 
 static int
 evict_writes (int fd, uint64_t offset, size_t len)
 {
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&window_fd_lock);
   struct write_window oldest = { 0 };
 
   {
@@ -138,7 +141,8 @@ evict_writes (int fd, uint64_t offset, size_t len)
      * (asynchronously).
      */
     if (sync_file_range (fd, offset, len, SYNC_FILE_RANGE_WRITE) == -1) {
-      nbdkit_error ("sync_file_range: cache=none: starting eviction: %m");
+      nbdkit_error ("sync_file_range: reduce-memory-pressure: "
+                    "starting eviction: %m");
       return -1;
     }
 
@@ -147,20 +151,28 @@ evict_writes (int fd, uint64_t offset, size_t len)
     window[NR_WINDOWS-1].offset = offset;
     window[NR_WINDOWS-1].len = len;
   }
-  /* Release lock here. */
+  /* Release window_lock here, but hold window_fd_lock since we don't
+   * want file_close -> remove_fd_from_window to close oldest.fd
+   */
 
   /* Evict the oldest window from the page cache (synchronously). */
   if (oldest.len > 0) {
+    int r;
+
     if (sync_file_range (oldest.fd, oldest.offset, oldest.len,
                          SYNC_FILE_RANGE_WAIT_BEFORE |
                          SYNC_FILE_RANGE_WRITE |
                          SYNC_FILE_RANGE_WAIT_AFTER) == -1) {
-      nbdkit_error ("sync_file_range: cache=none: evicting oldest window: %m");
+      nbdkit_error ("sync_file_range: reduce-memory-pressure: "
+                    "evicting oldest window: %m");
       return -1;
     }
-    if (posix_fadvise (oldest.fd, oldest.offset, oldest.len,
-                       POSIX_FADV_DONTNEED) == -1)
+    r = posix_fadvise (oldest.fd, oldest.offset, oldest.len,
+                       POSIX_FADV_DONTNEED);
+    if (r != 0) {
+      errno = r;
       nbdkit_debug ("posix_fadvise: POSIX_FADV_DONTNEED: (ignored): %m");
+    }
   }
 
   return 0;
@@ -172,7 +184,7 @@ evict_writes (int fd, uint64_t offset, size_t len)
 static void
 remove_fd_from_window (int fd)
 {
-  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&window_lock);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&window_fd_lock);
   size_t i;
 
   for (i = 0; i < NR_WINDOWS; ++i)
@@ -281,11 +293,20 @@ file_config (const char *key, const char *value)
       return -1;
     }
   }
+  else if (strcmp (key, "reduce-memory-pressure") == 0) {
+    int r;
+
+    r = nbdkit_parse_bool (value);
+    if (r == -1)
+      return -1;
+    reduce_memory_pressure = r;
+  }
+  /* For compatibility with nbdkit < 1.46 */
   else if (strcmp (key, "cache") == 0) {
     if (strcmp (value, "default") == 0)
-      cache_mode = cache_default;
+      reduce_memory_pressure = false;
     else if (strcmp (value, "none") == 0)
-      cache_mode = cache_none;
+      reduce_memory_pressure = true;
     else {
       nbdkit_error ("unknown cache mode: %s", value);
       return -1;
@@ -386,7 +407,7 @@ file_config_complete (void)
   "fd=<FILE_DESCRIPTOR>    Serve file attached to file descriptor\n" \
   "dir=<DIRNAME>           A directory containing files to serve\n" \
   "dirfd=<FILE_DESCRIPTOR> Serve dir attached to file descriptor\n" \
-  "cache=<MODE>            Set use of caching (default, none)\n" \
+  "reduce-memory-pressure=true  Reduce use of page cache (default: false)\n" \
   "fadvise=<LEVEL>         Set fadvise hint (normal, random, sequential)"
 
 /* Print some extra information about how the plugin was compiled. */
@@ -685,8 +706,10 @@ file_open (int readonly)
     /* This is a hint so we ignore failures. */
 #ifdef HAVE_POSIX_FADVISE
     int r = posix_fadvise (h->fd, 0, 0, fadvise_mode);
-    if (r == -1)
-      nbdkit_debug ("posix_fadvise: %s: %m (ignored)", h->name);
+    if (r != 0) {
+      errno = r;
+      nbdkit_debug ("posix_fadvise: %s: (ignored): %m", h->name);
+    }
 #else
     nbdkit_debug ("fadvise is not supported");
 #endif
@@ -932,7 +955,7 @@ file_pread (void *handle, void *buf, uint32_t count, uint64_t offset,
 
 #if defined (HAVE_POSIX_FADVISE) && defined (POSIX_FADV_DONTNEED)
   /* On Linux this will evict the pages we just read from the page cache. */
-  if (cache_mode == cache_none)
+  if (reduce_memory_pressure)
     posix_fadvise (h->fd, orig_offset, orig_count, POSIX_FADV_DONTNEED);
 #endif
 
@@ -966,7 +989,7 @@ file_pwrite (void *handle, const void *buf, uint32_t count, uint64_t offset,
     return -1;
 
 #if EVICT_WRITES
-  if (cache_mode == cache_none) {
+  if (reduce_memory_pressure) {
     if (evict_writes (h->fd, orig_offset, orig_count) == -1)
       return -1;
   }
